#!/usr/bin/python2.7

import sys
import argparse
import traceback
import errno
import sandglass

def parse_args(args=None):
	parser = argparse.ArgumentParser(prog="sgl", description="Sandglass CLI tool")
	subparsers = parser.add_subparsers(help="available commands")
	parser.add_argument("--debug", "-D", help="enable debugging", dest="debug", action="store_true")

	# status command

	parser_status = subparsers.add_parser("status", help="show information about directory state")
	parser_status.set_defaults(command=status)

	# stall command

	parser_stall = subparsers.add_parser("stall", help="temporarily disable autocommit for a package")
	parser_stall.set_defaults(command=stall)

	parser_stall.add_argument("--all", action="store_true", help="disable autocommit for all packages")
	parser_stall.add_argument("packages", metavar="package", nargs="*", help="package name")

	# auto command

	parser_auto = subparsers.add_parser("auto", help="re-enable autocommit for a package")
	parser_auto.set_defaults(command=auto)

	parser_auto.add_argument("--all", action="store_true", help="re-enable autocommit for all packages")
	parser_auto.add_argument("packages", metavar="package", nargs="*", help="package name")

	# init command

	parser_init = subparsers.add_parser("init", help="initialize Sandglass versioning")
	parser_init.set_defaults(command=init)

	parser_init.add_argument("--force", action="store_true", help="delete old history before initialization")
	parser_init.add_argument("--empty", action="store_true", help="do not import '/etc' into versioning system")

	# log command

	parser_log = subparsers.add_parser("log", help="show overall/package/file history")
	parser_log.set_defaults(command=log)

	parser_log.add_argument("files", metavar="file", nargs="*", help="File, directory, or a @package.")

	# add command

	parser_add = subparsers.add_parser("add", help="add directory, file, or package to be watched by Sandglass")
	parser_add.set_defaults(command=add)

	parser_add.add_argument("--everything", action="store_true", help="Add all files (including non-configuration "
			"files)")
	parser_add.add_argument("files", metavar="file", nargs="+", help="File, directory, or a @package.")
	parser_add.add_argument("-m", dest="message", metavar="message", nargs="?", help="Override default commit message.")

	# rm command

	parser_rm = subparsers.add_parser("remove", help="remove directory, file, or package from list of watched "
			"files")
	parser_rm.set_defaults(command=rm)

	parser_rm.add_argument("files", metavar="file", nargs="+", help="File, directory, or a @package.")
	parser_rm.add_argument("-m", dest="message", metavar="message", help="Override default commit message.")

	# commit command

	parse_commit = subparsers.add_parser("commit", help="Commit changes to the versioning system.")
	parse_commit.set_defaults(command=commit)

	parse_commit.add_argument("files", metavar="file", nargs="*", help="File, directory, or a @package.")
	parse_commit.add_argument("-m", dest="message", metavar="message", help="Override default commit message.")

	# show command

	parse_show = subparsers.add_parser("show", help="Show revision content.")
	parse_show.set_defaults(command=show)

	parse_show.add_argument("commit", metavar="commit", nargs="?", default="HEAD", help="Commit name. Default "
			"is the last one.")
	parse_show.add_argument("--files", metavar="file", nargs="*", help="Limit the output to the given "
			"filenames.")

	# diff command

	parse_diff = subparsers.add_parser("diff", help="Show difference between two commits.")
	parse_diff.set_defaults(command=diff)

	parse_diff.add_argument("commit_old", metavar="commit", nargs="?", help="Commit name, default is the last one.")
	parse_diff.add_argument("commit_new", metavar="commit", nargs="?", help="Commit name, default is current "
			"working tree.")
	parse_diff.add_argument("--files", metavar="file", nargs="?", help="Limit the output to the given "
			"filenames")

	# dump command

	parse_dump = subparsers.add_parser("dump", help="Shows the content of a file at a given commit.")
	parse_dump.set_defaults(command=dump)

	parse_dump.add_argument("commit", metavar="commit", nargs="?", help="Commit name, default is the last one.")
	parse_dump.add_argument("file", metavar="file", help="Filename of the file.")

	# recover command

	parse_recover = subparsers.add_parser("recover", help="Recover files from given commit.")
	parse_recover.set_defaults(command=recover)

	parse_recover.add_argument("commit", metavar="commit", nargs="?", default="HEAD", help="Commit name, default "
			"is the last one.")
	parse_recover_files = parse_recover.add_mutually_exclusive_group(required=True)
	parse_recover_files.add_argument("--all", action="store_true", help="Recover all files from the commit.")
	parse_recover_files.add_argument("--files", metavar="file", nargs="*", help="Filename of the file.")

	# remote command

	parse_remote = subparsers.add_parser("remote", help="Set remote Sandglass server for synchronization.")
	parse_remote.set_defaults(command=remote)

	parse_remote.add_argument("server", nargs="?", default=None, help="Remote server hostname.")

	# sync command

	parse_sync = subparsers.add_parser("sync", help="Perform synchronization with remote server.")
	parse_sync.set_defaults(command=sync)

	# shell (debug only)

	#parser_shell = subparsers.add_parser("shell")
	#parser_shell.set_defaults(command=shell)

	return parser.parse_args(args)

def _commit_message_files(prefix, files):
	if type(files) is not list or len(files) == 0:
		return "%s all" % prefix
	else:
		return "%s %s ..." % (prefix, files[0])

def init(args, sgl):
	sgl.init(args.force)
	print "Initialized Sandglass storage."
	if not args.empty:
		sgl.add(["/etc"])
		sgl.commit("initial import of /etc")
		print "Imported configuration within '/etc'."

def log(args, sgl):
	print sgl.log(files=args.files)

def status(args, sgl):
	changes = sgl.changed().by_package()
	autocommit_helper = sandglass.autocommit.get_instance()
	for package in changes:
		changeset = changes[package]
		time_auto = autocommit_helper.message(changeset, package)
		if package is not None:
			package = "package %s" % package
		else:
			package = "none of the packages"

		print "Changes in %s:" % package
		print
		for file in changeset.files():
			operations = ", ".join(map(lambda c: c.name(), changeset[file]))
			print "  %-19s %s" % (file, operations)
		print
		print "  %-19s %s" % ("first change:", sandglass.utils.time_relative(changeset.time_first_change()))
		print "  %-19s %s" % ("last change:", sandglass.utils.time_relative(changeset.time_last_change()))
		print "  %-19s %s" % ("autocommit:", time_auto)
		print

def stall(args, sgl):
	autocommit_helper = sandglass.autocommit.get_instance()
	if args.all:
		autocommit_helper.set_stall_all()
	else:
		for package in args.packages:
			autocommit_helper.set_stall(package)
	print "Autocommitting disabled. Use 'sgl auto' to re-enable."

def auto(args, sgl):
	autocommit_helper = sandglass.autocommit.get_instance()
	if args.all:
		autocommit_helper.set_auto_all()
	else:
		for package in args.packages:
			autocommit_helper.set_auto(package)
	print "Autocommitting re-enabled."

def add(args, sgl):
	sgl.add(args.files, not args.everything)
	if not args.message:
		msg = _commit_message_files("add", args.files)
		sgl.commit(msg)
	else:
		sgl.commit(args.message)

def rm(args, sgl):
	sgl.rm(args.files)
	if not args.message:
		msg = _commit_message_files("remove", args.files)
		sgl.commit(msg)
	else:
		sgl.commit(args.message)

def commit(args, sgl):
	sgl.add_changed(args.files)
	if not args.message:
		msg = _commit_message_files("changes", args.files)
		sgl.commit(msg)
	else:
		sgl.commit(args.message)

def show(args, sgl):
	output = sgl.show(args.commit, args.files)
	if output:
		print output

def diff(args, sgl):
	output = sgl.diff(args.commit_old, args.commit_new, files=args.files)
	if output:
		print output

def dump(args, sgl):
	print sgl.dump(args.file, args.commit), # no-end-line

def recover(args, sgl):
	sgl.recover(args.commit, files=args.files, all=args.all)
	print "Sucessfully recovered."

def remote(args, sgl):
	servers = sgl.remote(args.server)
	if len(servers) == 0:
		print "No remote server is set."
	else:
		if args.server is not None:
			print "Remote server updated. Use 'sgl sync' to perform the synchronization."
		for server in servers:
			print "Synchronization with remote server '%s' enabled." % servers[server]

def sync(args, sgl):
	sgl.sync()
	print "Synchronization is done."

# use for debugging
def shell(args, sgl):
	import subprocess
	subprocess.call(["sh"])

if __name__ == "__main__":
	args = parse_args()

	# convert path of all file arguments to normalized absolute path
	if "files" in args and type(args.files) is list:
		args.files = map(sandglass.utils.abs_path, args.files)
	if "file" in args and type(args.file) is str:
		args.file = sandglass.utils.abs_path(args.file)

	sgl = sandglass.Sandglass()

	try:
		args.command(args, sgl)
	except IOError as e:
		if e.errno == errno.EPIPE:
			pass
		else:
			print >> sys.stderr, "ERROR: %s" % e
	except Exception as e:
		print >> sys.stderr, "ERROR: %s" % e
		if args.debug:
			traceback.print_exc(sys.stderr)
		sys.exit(1)

	sys.exit(0)
